from .case import Case
from .case_listener import CaseListener
from .case_result import CaseResult
from .case_runner import CaseRunner
from .scenario import Scenario
from .scenario_result import ScenarioResult
from .scenario_runner import ScenarioRunner
from .scenario_task import ScenarioTask

__all__ = [
    'Case',
    'CaseRunner',
    'CaseListener',
    'CaseResult',
    'Scenario',
    'ScenarioRunner',
    'ScenarioResult',
    'ScenarioTask',
]



from preacher.compilation.extraction.extraction import ExtractionCompiler, add_default_extractions
from preacher.compilation.verification.matcher import MatcherFactoryCompiler
from preacher.compilation.verification.matcher import add_default_matchers
from . import hookimpl


@hookimpl(tryfirst=True)
def preacher_add_matchers(compiler: MatcherFactoryCompiler) -> None:
    add_default_matchers(compiler)


@hookimpl(tryfirst=True)
def preacher_add_extractions(compiler: ExtractionCompiler) -> None:
    add_default_extractions(compiler)



"""Date and time compilation."""

import re
from datetime import timedelta

from preacher.core.datetime import DatetimeFormat, ISO8601, StrftimeFormat
from .error import CompilationError
from .util.type import ensure_str, ensure_optional_str

TIMEDELTA_PATTERN = re.compile(r'([+\-]?\d+)\s*(day|hour|minute|second)s?')


def compile_datetime_format(obj: object) -> DatetimeFormat:
    """
    Args:
        obj: The compiled value, which should be `None` or a string.
    Raises:
        CompilationError: When compilation fails.
    """
    format_string = ensure_optional_str(obj)
    if format_string is None or format_string.lower() == 'iso8601':
        return ISO8601
    return StrftimeFormat(format_string)


def compile_timedelta(obj: object) -> timedelta:
    """
    Args:
        obj: The compiled value, which should be a string.
    Raises:
        CompilationError: When compilation fails.
    """
    obj = ensure_str(obj)
    normalized = obj.strip().lower()
    if not normalized or normalized == 'now':
        return timedelta()

    match = TIMEDELTA_PATTERN.match(normalized)
    if not match:
        raise CompilationError(f'Invalid timedelta format: {obj}')
    offset = int(match.group(1))
    unit = match.group(2) + 's'
    return timedelta(**{unit: offset})



"""Extraction package."""

from preacher.core.extraction.analysis import Analyzer
from preacher.core.extraction.analysis import MappingAnalyzer
from preacher.core.extraction.analysis import ResponseBodyAnalyzer
from preacher.core.extraction.analysis import analyze_data_obj
from preacher.core.extraction.error import ExtractionError
from preacher.core.extraction.extraction import Extractor

__all__ = [
    'Analyzer',
    'ResponseBodyAnalyzer',
    'MappingAnalyzer',
    'analyze_data_obj',
    'ExtractionError',
    'Extractor',
]



"""
Value interpretation.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Generic, Optional, TypeVar

from preacher.core.datetime import DatetimeWithFormat, DatetimeFormat, ISO8601, now

T = TypeVar('T')


@dataclass(frozen=True)
class ValueContext:
    origin_datetime: Optional[datetime] = None


class Value(ABC, Generic[T]):

    @abstractmethod
    def resolve(self, context: Optional[ValueContext] = None) -> T:
        raise NotImplementedError()


class StaticValue(Value[T]):

    def __init__(self, value: T):
        self._value = value

    def resolve(self, context: Optional[ValueContext] = None) -> T:
        return self._value


class RelativeDatetime(Value[DatetimeWithFormat]):

    def __init__(
        self,
        delta: Optional[timedelta] = None,
        fmt: Optional[DatetimeFormat] = None,
    ):
        self._delta = delta or timedelta()
        self._fmt = fmt or ISO8601

    def resolve(
        self,
        context: Optional[ValueContext] = None,
    ) -> DatetimeWithFormat:
        if not context:
            context = ValueContext()

        origin = context.origin_datetime or now()
        return DatetimeWithFormat(origin + self._delta, self._fmt)



"""
Response descriptions, which verify the status code, the headers
and the body.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from preacher.core.extraction import ResponseBodyAnalyzer, MappingAnalyzer
from preacher.core.request import Response
from preacher.core.status import Status, Statused, merge_statuses
from preacher.core.value import ValueContext
from .description import Description
from .predicate import Predicate
from .verification import Verification


@dataclass(frozen=True)
class ResponseVerification(Statused):
    response_id: str
    status_code: Verification = field(default_factory=Verification)
    headers: Verification = field(default_factory=Verification)
    body: Verification = field(default_factory=Verification)

    @property  # HACK should be cached.
    def status(self) -> Status:
        return merge_statuses([self.status_code.status, self.headers.status, self.body.status])


class ResponseDescription:

    def __init__(
        self,
        status_code: Optional[List[Predicate]] = None,
        headers: Optional[List[Description]] = None,
        body: Optional[List[Description]] = None,
    ):
        self._status_code = status_code or []
        self._headers = headers or []
        self._body = body or []

    def verify(
        self,
        response: Response,
        context: Optional[ValueContext] = None,
    ) -> ResponseVerification:
        status_code = Verification.collect(
            p.verify(response.status_code, context) for p in self._status_code
        )

        header_analyzer = MappingAnalyzer(response.headers)
        headers = Verification.collect(
            d.verify(header_analyzer, context) for d in self._headers
        )

        body_analyzer = ResponseBodyAnalyzer(response.body)
        body = Verification.collect(
            d.verify(body_analyzer, context) for d in self._body
        )

        return ResponseVerification(
            response_id=response.id,
            status_code=status_code,
            headers=headers,
            body=body,
        )



"""Scenario compilation."""

from functools import partial
from typing import Iterator, List, Optional, Mapping

from preacher.compilation.argument import Arguments, inject_arguments
from preacher.compilation.error import on_key
from preacher.compilation.parameter import Parameter, compile_parameter
from preacher.compilation.util.functional import (
    map_compile,
    compile_flattening,
)
from preacher.compilation.util.type import (
    ensure_bool,
    ensure_optional_str,
    ensure_list,
    ensure_mapping,
)
from preacher.compilation.verification import DescriptionCompiler
from preacher.core.scenario import Scenario, Case
from .case import CaseCompiler

_KEY_LABEL = 'label'
_KEY_WHEN = 'when'
_KEY_DEFAULT = 'default'
_KEY_ORDERED = 'ordered'
_KEY_CASES = 'cases'
_KEY_PARAMETERS = 'parameters'
_KEY_SUBSCENARIOS = 'subscenarios'


class ScenarioCompiler:

    def __init__(self, description: DescriptionCompiler, case: CaseCompiler):
        self._description = description
        self._case = case

    def compile(self, obj: object, arguments: Optional[Arguments] = None) -> Scenario:
        """
        Compile the given object into a scenario.

        Args:
            obj: A compiled object, which should be a mapping.
            arguments: Arguments to inject.
        Returns:
            The scenario as the result of compilation.
        Raises:
            CompilationError: when the compilation fails.
        """

        obj = ensure_mapping(obj)
        arguments = arguments or {}

        label_obj = inject_arguments(obj.get(_KEY_LABEL), arguments)
        with on_key(_KEY_LABEL):
            label = ensure_optional_str(label_obj)

        parameters_obj = obj.get(_KEY_PARAMETERS)
        if parameters_obj is not None:
            with on_key(_KEY_PARAMETERS):
                parameters_obj = ensure_list(parameters_obj)
                parameters = list(
                    map_compile(compile_parameter, parameters_obj)
                )
            subscenarios = [
                self._compile_parameterized(obj, arguments, parameter)
                for parameter in parameters
            ]
            return Scenario(label=label, subscenarios=subscenarios)

        ordered_obj = inject_arguments(obj.get(_KEY_ORDERED, True), arguments)
        with on_key(_KEY_ORDERED):
            ordered = ensure_bool(ordered_obj)

        default_obj = inject_arguments(obj.get(_KEY_DEFAULT, {}), arguments)
        with on_key(_KEY_DEFAULT):
            case_compiler = self._case.compile_default(default_obj)

        condition_obj = inject_arguments(obj.get(_KEY_WHEN, []), arguments)
        with on_key(_KEY_WHEN):
            conditions = self._compile_conditions(condition_obj)

        case_obj = inject_arguments(obj.get(_KEY_CASES, []), arguments)
        with on_key(_KEY_CASES):
            cases = self._compile_cases(case_compiler, case_obj)

        subscenario_obj = obj.get(_KEY_SUBSCENARIOS, [])
        with on_key(_KEY_SUBSCENARIOS):
            subscenarios = self._compile_subscenarios(
                case_compiler,
                subscenario_obj,
                arguments,
            )

        return Scenario(
            label=label,
            ordered=ordered,
            conditions=conditions,
            cases=cases,
            subscenarios=subscenarios,
        )

    def compile_flattening(
        self,
        obj: object,
        arguments: Optional[Arguments] = None,
    ) -> Iterator[Scenario]:
        """
        Compile the given object into a scenario with flattening:
        a nested object list results in a flattened scenario.

        Args:
            obj: A compiled object or a list.
            arguments: Arguments to inject.
        Returns:
            A scenario iterator as the result of compilation.
        Raises:
            CompilationError: when the compilation fails for each iteration.
        """

        compile = partial(self.compile, arguments=arguments)
        return compile_flattening(compile, obj)

    def _compile_conditions(self, obj: object):
        return list(map_compile(self._description.compile, ensure_list(obj)))

    @staticmethod
    def _compile_cases(case_compiler: CaseCompiler, obj: object) -> List[Case]:
        return list(
            map_compile(case_compiler.compile_fixed, ensure_list(obj))
        )

    def _compile_subscenarios(
        self,
        case: CaseCompiler,
        obj: object,
        arguments: Arguments,
    ) -> List[Scenario]:
        compiler = ScenarioCompiler(description=self._description, case=case)
        return list(map_compile(
            lambda sub_obj: compiler.compile(sub_obj, arguments=arguments),
            ensure_list(obj),
        ))

    def _compile_parameterized(
        self,
        obj: Mapping,
        arguments: Arguments,
        parameter: Parameter,
    ) -> Scenario:
        template = {
            k: v for (k, v) in obj.items()
            if k not in (_KEY_LABEL, _KEY_PARAMETERS)
        }
        template['label'] = parameter.label

        arguments = dict(arguments)
        arguments.update(parameter.arguments)
        return self.compile(template, arguments)



import contextlib
import io
import logging
from typing import Iterator

from preacher.core.request import ExecutionReport
from preacher.core.scenario import ScenarioResult, CaseResult
from preacher.core.status import Status
from preacher.core.verification import ResponseVerification, Verification

_LEVEL_MAP = {
    Status.SKIPPED: logging.DEBUG,
    Status.SUCCESS: logging.INFO,
    Status.UNSTABLE: logging.WARN,
    Status.FAILURE: logging.ERROR,
}


class LoggingReporter:

    def __init__(self, logger: logging.Logger):
        self._logger = logger
        self._indent = ''

    def show_scenario_result(self, scenario: ScenarioResult) -> None:
        status = scenario.status
        level = _LEVEL_MAP[status]

        label = scenario.label or 'Not labeled scenario'
        self._log(level, '%s: %s', label, status)

        message = scenario.message
        if message:
            with self._nesting():
                self._multi_line_message(level, message)

        with self._nesting():
            for case in scenario.cases.items:
                self.show_case_result(case)

            for subscenario in scenario.subscenarios.items:
                self.show_scenario_result(subscenario)

    def show_case_result(self, case: CaseResult) -> None:
        status = case.status
        level = _LEVEL_MAP[status]

        label = case.label or 'Not labeled case'
        self._log(level, '%s: %s', label, status)
        with self._nesting():
            self.show_execution(case.execution)

            response = case.response
            if response:
                self.show_response_verification(response)

    def show_execution(self, execution: ExecutionReport) -> None:
        status = execution.status
        level = _LEVEL_MAP[status]

        self._log(level, 'Execution: %s', status)
        if execution.message:
            with self._nesting():
                self._multi_line_message(level, execution.message)

    def show_response_verification(
        self,
        verification: ResponseVerification,
        label: str = 'Response',
    ) -> None:
        status = verification.status
        level = _LEVEL_MAP[status]
        self._log(level, '%s: %s', label, status)
        with self._nesting():
            self.show_verification(
                verification=verification.status_code,
                label='Status Code',
            )
            self.show_verification(
                verification=verification.headers,
                label='Headers',
                child_label='Description',
            )
            self.show_verification(
                verification=verification.body,
                label='Body',
                child_label='Description',
            )

    def show_verification(
        self,
        verification: Verification,
        label: str,
        child_label: str = 'Predicate',
    ) -> None:
        status = verification.status
        level = _LEVEL_MAP[status]
        self._log(level, '%s: %s', label, status)
        message = verification.message
        if message:
            with self._nesting():
                self._multi_line_message(level, message)

        with self._nesting():
            for idx, child in enumerate(verification.children):
                self.show_verification(child, f'{child_label} {idx + 1}')

    def show_status(self, status: Status) -> None:
        level = _LEVEL_MAP[status]
        self._log(level, '%s', status)

    def _log(self, level: int, message: str, *args) -> None:
        self._logger.log(level, self._indent + message, *args)

    def _multi_line_message(self, level: int, message: str) -> None:
        for line in io.StringIO(message):
            self._log(level, line.rstrip())

    @contextlib.contextmanager
    def _nesting(self) -> Iterator[None]:
        original = self._indent
        self._indent += '..'
        yield
        self._indent = original



"""URL parameters compilation."""

from datetime import date
from typing import Mapping, Optional

from preacher.compilation.argument import Arguments, inject_arguments
from preacher.compilation.error import CompilationError, on_key
from preacher.compilation.util.functional import map_compile
from preacher.core.request import UrlParams, UrlParam, UrlParamValue
from preacher.core.value import RelativeDatetime


def compile_url_param_value(value: object) -> UrlParamValue:
    if value is None:
        return value
    if isinstance(value, bool):
        return value
    if isinstance(value, int):
        return value
    if isinstance(value, float):
        return value
    if isinstance(value, str):
        return value
    if isinstance(value, date):
        return value
    if isinstance(value, RelativeDatetime):  # HACK: relax typing.
        return value
    raise CompilationError(
        f'Not allowed type for a request parameter value: {value.__class__}'
    )


def compile_url_param(value: object) -> UrlParam:
    if value is None:
        return value
    if isinstance(value, list):
        return list(map_compile(compile_url_param_value, value))
    try:
        return compile_url_param_value(value)
    except CompilationError as error:
        raise CompilationError(
            f'Not allowed type for a request parameter: {value.__class__}',
            cause=error
        )


def compile_url_params(
    obj: object,
    arguments: Optional[Arguments] = None,
) -> UrlParams:
    """
    Compiles an object into URL parameters.

    Args:
        obj: A compiled object, which should be a mapping or a string.
        arguments: Arguments to inject.
    Returns:
        The result of compilation.
    Raises:
        CompilationError: when compilation fails.
    """
    obj = inject_arguments(obj, arguments)

    if isinstance(obj, str):
        return obj

    if not isinstance(obj, Mapping):
        raise CompilationError('Must be a mapping or a string')

    compiled = {}
    for key, value in obj.items():
        assert isinstance(key, str)  # Satisfied in injecting arguments.
        with on_key(key):
            compiled[key] = compile_url_param(value)
    return compiled



"""
Descriptions, which extract a value from an analyzer and test
along the given predicates.
"""

from typing import List, Optional

from preacher.core.extraction import Analyzer, Extractor
from preacher.core.value import ValueContext
from .predicate import Predicate
from .verification import Verification


class Description:

    def __init__(self, extractor: Extractor, predicates: List[Predicate]):
        self._extractor = extractor
        self._predicates = predicates

    def verify(self, analyzer: Analyzer, context: Optional[ValueContext] = None) -> Verification:
        try:
            verified_value = self._extractor.extract(analyzer)
        except Exception as error:
            return Verification.of_error(error)

        return Verification.collect(
            predicate.verify(verified_value, context)
            for predicate in self._predicates
        )



from collections.abc import Mapping
from typing import Optional

from .util.functional import run_recursively

Arguments = Mapping


class Argument:

    def __init__(self, key: str):
        self._key = key

    @property
    def key(self) -> str:
        return self._key

    def apply_arguments(self, arguments: Optional[Arguments] = None) -> object:
        arguments = arguments or {}
        return arguments.get(self._key)


def _inject_arguments(obj: object, arguments: Optional[Arguments]) -> object:
    if isinstance(obj, Argument):
        return obj.apply_arguments(arguments)
    return obj


def inject_arguments(
    obj: object,
    arguments: Optional[Arguments] = None,
) -> object:
    return run_recursively(lambda o: _inject_arguments(o, arguments), obj)



"""
Predicates, which tests a given value.
"""

from abc import ABC, abstractmethod
from typing import Optional

from preacher.core.value import ValueContext
from .verification import Verification


class Predicate(ABC):
    """Predicate interface."""

    @abstractmethod
    def verify(self, actual: object, context: Optional[ValueContext] = None) -> Verification:
        raise NotImplementedError()



"""Matcher compilation."""

from collections.abc import Mapping
from datetime import datetime, timezone
from typing import Callable, Dict, Iterable, Tuple, Union

import hamcrest
from hamcrest.core.matcher import Matcher

from preacher.compilation.datetime import compile_timedelta
from preacher.compilation.error import CompilationError, on_key
from preacher.compilation.util.functional import map_compile
from preacher.compilation.util.type import ensure_list
from preacher.core.datetime import DatetimeWithFormat
from preacher.core.value import Value, StaticValue, RelativeDatetime
from preacher.core.verification.hamcrest import before, after
from preacher.core.verification.matcher import MatcherFactory
from preacher.core.verification.matcher import MatcherFunc
from preacher.core.verification.matcher import RecursiveMatcherFactory
from preacher.core.verification.matcher import StaticMatcherFactory
from preacher.core.verification.matcher import ValueMatcherFactory
from preacher.core.verification.type import require_type

ValueFunc = Callable[[object], Value]


class MatcherFactoryCompiler:
    """
    A matcher factory compiler.
    """

    _DEFAULT_VALUE_FUNC = StaticValue

    def __init__(self):
        self._static: Dict[str, MatcherFactory] = {}
        self._taking_value: Dict[str, Tuple[MatcherFunc, ValueFunc]] = {}
        self._recursive: Dict[str, Tuple[MatcherFunc, bool]] = {}

    def add_static(
        self,
        keys: Union[str, Iterable[str]],
        item: Union[MatcherFactory, Matcher],
    ) -> None:
        """
        Add a static matcher on given keys.

        Args:
            keys: The key(s) of the matcher.
            item: A matcher or a ``MatcherFactory`` to assign.
        """

        if isinstance(item, Matcher):
            item = StaticMatcherFactory(item)
        for key in self._ensure_keys(keys):
            self._static[key] = item

    def add_taking_value(
        self,
        keys: Union[str, Iterable[str]],
        matcher_func: MatcherFunc,
        value_func: ValueFunc = _DEFAULT_VALUE_FUNC,
    ) -> None:
        """
        Add a matcher taking a value on given keys.

        Args:
            keys: The key(s) of the matcher.
            matcher_func: A function that takes an object and returns a matcher.
            value_func (optional): A function that takes an object and returns a ``Value`` object.
        """

        for key in self._ensure_keys(keys):
            self._taking_value[key] = (matcher_func, value_func)

    def add_recursive(
        self,
        keys: Union[str, Iterable[str]],
        matcher_func: MatcherFunc,
        multiple: bool = True,
    ) -> None:
        """
        Add a matcher taking one or more matchers.

        Args:
            keys: The key(s) of the matcher.
            matcher_func: A function that takes one or more matchers and returns a matcher.
            multiple: Whether the matcher can take multiple arguments or not.
        """

        for key in self._ensure_keys(keys):
            self._recursive[key] = (matcher_func, multiple)

    def compile(self, obj: object) -> MatcherFactory:
        """
        Compile an object into a matcher factory.

        Args:
            obj: A compiled object.
        Returns:
            The result of compilation.
        Raises:
            CompilationError: when compilation fails.
        """

        if isinstance(obj, str) and obj in self._static:
            return self._static[obj]

        if isinstance(obj, Mapping):
            if len(obj) != 1:
                message = f'Must have only 1 element, but has {len(obj)}'
                raise CompilationError(message)

            key, value_obj = next(iter(obj.items()))
            if key in self._taking_value:
                return self._compile_taking_value(key, value_obj)
            if key in self._recursive:
                return self._compile_recursive(key, value_obj)

        return ValueMatcherFactory(hamcrest.equal_to, StaticValue(obj))

    def _compile_taking_value(self, key: str, obj: object):
        matcher_func, value_func = self._taking_value[key]
        with on_key(key):
            value = value_func(obj)
        return ValueMatcherFactory(matcher_func, value)

    def _compile_recursive(self, key: str, obj: object):
        matcher_func, multiple = self._recursive[key]
        if multiple:
            objs = ensure_list(obj)
            inner_matchers = list(map_compile(self.compile, objs))
        else:
            with on_key(key):
                inner_matchers = [self.compile(obj)]
        return RecursiveMatcherFactory(matcher_func, inner_matchers)

    @staticmethod
    def _ensure_keys(keys: Union[str, Iterable[str]]) -> Iterable[str]:
        if isinstance(keys, str):
            return (keys,)
        return keys


def add_default_matchers(compiler: MatcherFactoryCompiler) -> None:
    """
    Add default matchers to a compiler.

    Args:
        compiler: A compiler to be modified.
    """

    compiler.add_recursive(('be',), hamcrest.is_, multiple=False)

    # For objects.
    compiler.add_static(('be_null',), hamcrest.none())
    compiler.add_static(('not_be_null',), hamcrest.not_none())
    compiler.add_taking_value(('equal',), hamcrest.equal_to)
    compiler.add_recursive(('have_length',), hamcrest.has_length, multiple=False)

    # For comparable values.
    compiler.add_taking_value(('be_greater_than',), hamcrest.greater_than)
    compiler.add_taking_value(('be_greater_than_or_equal_to',), hamcrest.greater_than_or_equal_to)
    compiler.add_taking_value(('be_less_than',), hamcrest.less_than)
    compiler.add_taking_value(('be_less_than_or_equal_to',), hamcrest.less_than_or_equal_to)

    # For strings.
    compiler.add_taking_value(('contain_string',), require_type(str, hamcrest.contains_string))
    compiler.add_taking_value(('start_with',), require_type(str, hamcrest.starts_with))
    compiler.add_taking_value(('end_with',), require_type(str, hamcrest.ends_with))
    compiler.add_taking_value(('match_regexp',), require_type(str, hamcrest.matches_regexp))

    # For collections.
    compiler.add_recursive(('have_item',), hamcrest.has_item, multiple=False)
    compiler.add_recursive(('have_items',), hamcrest.has_items)
    compiler.add_recursive(('contain_exactly',), hamcrest.contains_exactly)
    compiler.add_recursive(('contain_in_any_order',), hamcrest.contains_inanyorder)

    # For datetime.
    compiler.add_taking_value(('be_before',), before, _compile_datetime_value)
    compiler.add_taking_value(('be_after',), after, _compile_datetime_value)

    # For collections.
    compiler.add_static(('be_empty',), StaticMatcherFactory(hamcrest.empty()))

    # Logical.
    compiler.add_static('anything', StaticMatcherFactory(hamcrest.anything()))
    compiler.add_recursive(('not',), hamcrest.not_, multiple=False)
    compiler.add_recursive(('all_of',), hamcrest.all_of)
    compiler.add_recursive(('any_of',), hamcrest.any_of)


def _compile_datetime_value(value: object) -> Value[DatetimeWithFormat]:
    if isinstance(value, datetime):
        if not value.tzinfo:
            value = value.replace(tzinfo=timezone.utc)
        return StaticValue(DatetimeWithFormat(value))

    delta = compile_timedelta(value)
    return RelativeDatetime(delta)



"""Extraction."""

from abc import ABC, abstractmethod
from typing import TypeVar

from .analysis import Analyzer

T = TypeVar('T')


class Extractor(ABC):

    @abstractmethod
    def extract(self, analyzer: Analyzer) -> object:
        """
        Raises:
            EvaluationError: when the evaluation of this extractor fails.
        """
        raise NotImplementedError()



from pluggy.hooks import HookspecMarker

from preacher.compilation.extraction import ExtractionCompiler
from preacher.compilation.verification.matcher import MatcherFactoryCompiler

hookspec = HookspecMarker('preacher')


@hookspec
def preacher_add_matchers(compiler: MatcherFactoryCompiler) -> None:
    """
    Add Hamcrest matchers to a compiler.

    Args:
        compiler: A compiler to modify.
    """


@hookspec
def preacher_add_extractions(compiler: ExtractionCompiler) -> None:
    """
    Add extractions to a compiler

    Args:
        compiler: A compiler to modify.
    """



from datetime import timedelta
from typing import Optional

from yaml import BaseLoader, MappingNode, ScalarNode, Node

from preacher.compilation.datetime import compile_timedelta, compile_datetime_format
from preacher.core.datetime import DatetimeFormat
from preacher.core.value import RelativeDatetime
from .error import YamlError, on_node

_KEY_DELTA = 'delta'
_KEY_FORMAT = 'format'


def construct_relative_datetime(loader: BaseLoader, node: Node) -> RelativeDatetime:
    if isinstance(node, ScalarNode):
        return _construct_relative_datetime_of_scalar(loader, node)
    elif isinstance(node, MappingNode):
        return _construct_relative_datetime_of_mapping(loader, node)
    else:
        message = 'Invalid relative datetime value format'
        raise YamlError(message, mark=node.start_mark)


def _construct_relative_datetime_of_scalar(
    loader: BaseLoader,
    node: ScalarNode,
) -> RelativeDatetime:
    obj = loader.construct_scalar(node)
    with on_node(node):
        delta = compile_timedelta(obj)
    return RelativeDatetime(delta)


def _construct_relative_datetime_of_mapping(
    loader: BaseLoader,
    node: MappingNode,
) -> RelativeDatetime:
    delta: Optional[timedelta] = None
    format: Optional[DatetimeFormat] = None

    for key_node, value_node in node.value:
        if key_node.value == _KEY_DELTA:
            obj = loader.construct_scalar(value_node)
            with on_node(value_node):
                delta = compile_timedelta(obj)
            continue
        if key_node.value == _KEY_FORMAT:
            obj = loader.construct_scalar(value_node)
            with on_node(node):
                format = compile_datetime_format(obj)
            continue

    return RelativeDatetime(delta, format)



"""Preacher plugin definitions."""

from pluggy import PluginManager

from . import hookspec
from . import impl

__all__ = ['get_plugin_manager']


def get_plugin_manager() -> PluginManager:
    """
    Get a plugin manager that the entry points and the default implementation are already loaded.
    """
    manager = PluginManager('preacher')
    manager.add_hookspecs(hookspec)
    manager.load_setuptools_entrypoints('preacher')
    manager.register(impl)
    return manager



"""Matchers."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Generic, List, Optional, TypeVar

from hamcrest import assert_that
from hamcrest.core.matcher import Matcher

from preacher.core.status import Status
from preacher.core.value import Value, ValueContext
from .predicate import Predicate
from .verification import Verification

T = TypeVar('T')
MatcherFunc = Callable[..., Matcher]


class MatcherWrappingPredicate(Predicate):
    """Matcher implemented by hamcrest matchers."""

    def __init__(self, factory: MatcherFactory):
        self._factory = factory

    def verify(self, actual: object, context: Optional[ValueContext] = None) -> Verification:
        try:
            hamcrest_matcher = self._factory.create(context)
            assert_that(actual, hamcrest_matcher)
        except AssertionError as error:
            message = str(error).strip()
            return Verification(status=Status.UNSTABLE, message=message)
        except Exception as error:
            return Verification.of_error(error)

        return Verification.succeed()


class MatcherFactory(ABC):

    @abstractmethod
    def create(self, context: Optional[ValueContext] = None) -> Matcher:
        raise NotImplementedError()


class StaticMatcherFactory(MatcherFactory):

    def __init__(self, matcher: Matcher):
        self._matcher = matcher

    def create(self, context: Optional[ValueContext] = None) -> Matcher:
        return self._matcher


class ValueMatcherFactory(MatcherFactory, Generic[T]):

    def __init__(self, matcher_func: MatcherFunc, value: Value[T]):
        self._inner_factory = matcher_func
        self._value = value

    def create(self, context: Optional[ValueContext] = None) -> Matcher:
        resolved_value = self._value.resolve(context)
        return self._inner_factory(resolved_value)


class RecursiveMatcherFactory(MatcherFactory):

    def __init__(self, matcher_func: MatcherFunc, inner_factories: List[MatcherFactory]):
        self._matcher_func = matcher_func
        self._inner_factories = inner_factories

    def create(self, context: Optional[ValueContext] = None) -> Matcher:
        inner_matchers = (factory.create(context) for factory in self._inner_factories)
        return self._matcher_func(*inner_matchers)



"""Description compilation."""

from preacher.compilation.error import on_key
from preacher.compilation.extraction import ExtractionCompiler
from preacher.compilation.util.functional import map_compile
from preacher.compilation.util.type import ensure_list, ensure_mapping
from preacher.core.verification import Description
from .predicate import PredicateCompiler

_KEY_DESCRIBE = 'describe'
_KEY_SHOULD = 'should'


class DescriptionCompiler:

    def __init__(
        self,
        extraction: ExtractionCompiler,
        predicate: PredicateCompiler,
    ):
        self._extraction = extraction
        self._predicate = predicate

    def compile(self, obj: object):
        """`obj` should be a mapping."""

        obj = ensure_mapping(obj)

        extraction_obj = obj.get(_KEY_DESCRIBE)
        with on_key(_KEY_DESCRIBE):
            extractor = self._extraction.compile(extraction_obj)

        predicate_objs = ensure_list(obj.get(_KEY_SHOULD, []))
        with on_key(_KEY_SHOULD):
            predicates = list(map_compile(
                self._predicate.compile,
                predicate_objs,
            ))

        return Description(extractor=extractor, predicates=predicates)



"""Verification."""

from .description import Description
from .matcher import MatcherFactory
from .matcher import MatcherFunc
from .matcher import MatcherWrappingPredicate
from .matcher import RecursiveMatcherFactory
from .matcher import StaticMatcherFactory
from .matcher import ValueMatcherFactory
from .predicate import Predicate
from .response import ResponseDescription
from .response import ResponseVerification
from .type import require_type
from .verification import Verification

__all__ = [
    'Description',
    'MatcherWrappingPredicate',
    'MatcherFactory',
    'StaticMatcherFactory',
    'ValueMatcherFactory',
    'RecursiveMatcherFactory',
    'MatcherFunc',
    'Predicate',
    'ResponseDescription',
    'ResponseVerification',
    'require_type',
    'Verification',
]



import json
from abc import ABC, abstractmethod
from typing import Any, Optional

from preacher.core.util.functional import recursive_map
from preacher.core.util.serialization import to_serializable
from preacher.core.value import Value, ValueContext
from .url_param import UrlParams, resolve_url_params


class RequestBody(ABC):

    @property
    @abstractmethod
    def content_type(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def resolve(self, context: Optional[ValueContext] = None) -> Any:
        raise NotImplementedError()


class UrlencodedRequestBody(RequestBody):

    def __init__(self, params: UrlParams):
        self._params = params

    @property
    def content_type(self) -> str:
        return 'application/x-www-form-urlencoded'

    def resolve(self, context: Optional[ValueContext] = None) -> Any:
        return resolve_url_params(self._params, context)


class JsonRequestBody(RequestBody):

    def __init__(self, data: object):
        self._data = data

    @property
    def content_type(self) -> str:
        return 'application/json'

    def resolve(self, context: Optional[ValueContext] = None) -> Any:
        def _resolve_value(obj: object) -> object:
            if isinstance(obj, Value):
                return recursive_map(
                    _resolve_value,
                    _resolve_value(obj.resolve(context)),
                )
            return to_serializable(obj)

        resolved = recursive_map(_resolve_value, self._data)
        return json.dumps(resolved, separators=(',', ':'))



"""Extraction compilation."""

from collections.abc import Mapping
from functools import partial
from typing import Any, Callable, Optional

from preacher.compilation.error import CompilationError, on_key
from preacher.compilation.util.type import ensure_bool, ensure_str
from preacher.core.extraction.extraction import Extractor
from preacher.core.extraction.impl.key import KeyExtractor
from preacher.core.extraction.impl.jq_ import JqExtractor
from preacher.core.extraction.impl.jq_engine import PyJqEngine
from preacher.core.extraction.impl.xpath import XPathExtractor

_CAST_FUNC_MAP = {
    'int': int,
    'float': float,
    'string': str,
}
_KEY_JQ = 'jq'
_KEY_XPATH = 'xpath'
_KEY_KEY = 'key'
_KEY_MULTIPLE = 'multiple'
_KEY_CAST_TO = 'cast_to'


class ExtractionCompiler:

    def __init__(self):
        self._factory_map = {}

    def add_factory(self, key: str, factory):
        self._factory_map[key] = factory

    def compile(self, obj: object) -> Extractor:
        """`obj` should be a mapping or a string."""

        if isinstance(obj, str):
            return self.compile({_KEY_JQ: obj})

        if not isinstance(obj, Mapping):
            message = f'Must be a map or a string, given {type(obj)}'
            raise CompilationError(message)

        key_candidates = self._factory_map.keys() & obj.keys()
        key_count = len(key_candidates)
        if key_count != 1:
            raise CompilationError(f'Must have only 1 extraction key, but has {key_count}')
        factory_key = next(iter(key_candidates))
        factory = self._factory_map[factory_key]
        query = obj[factory_key]

        multiple_obj = obj.get(_KEY_MULTIPLE, False)
        with on_key(_KEY_MULTIPLE):
            multiple = ensure_bool(multiple_obj)

        cast: Optional[Callable[[object], Any]] = None
        cast_obj = obj.get(_KEY_CAST_TO)
        if cast_obj is not None:
            with on_key(_KEY_CAST_TO):
                cast = self._compile_cast(cast_obj)

        return factory(query, multiple=multiple, cast=cast)

    @staticmethod
    def _compile_cast(obj: object) -> Callable[[object], Any]:
        """`obj` should be a string."""

        key = ensure_str(obj)
        cast = _CAST_FUNC_MAP.get(key)
        if not cast:
            raise CompilationError(f'Invalid value: {key}')

        return cast


def add_default_extractions(compiler: ExtractionCompiler) -> None:
    if PyJqEngine.is_available():
        compiler.add_factory(_KEY_JQ, partial(JqExtractor, PyJqEngine()))
    compiler.add_factory(_KEY_XPATH, XPathExtractor)
    compiler.add_factory(_KEY_KEY, KeyExtractor)



from concurrent.futures import Executor
from dataclasses import dataclass, field
from datetime import datetime
from typing import Callable

from preacher.core.datetime import now
from preacher.core.extraction import analyze_data_obj
from preacher.core.status import Status
from preacher.core.value import ValueContext
from preacher.core.verification import Verification
from .case_runner import CaseRunner
from .scenario import Scenario
from .scenario_result import ScenarioResult
from .scenario_task import ScenarioTask, StaticScenarioTask, RunningScenarioTask
from .util.concurrency import OrderedCasesTask, UnorderedCasesTask


@dataclass(frozen=True)
class ScenarioContext:
    starts: datetime = field(default_factory=now)
    base_url: str = ''


class ScenarioRunner:

    def __init__(self, executor: Executor, case_runner: CaseRunner):
        self._executor = executor
        self._case_runner = case_runner

    def submit(self, scenario: Scenario) -> ScenarioTask:
        context = ScenarioContext(base_url=self._case_runner.base_url)
        context_analyzer = analyze_data_obj(context)
        value_context = ValueContext(origin_datetime=context.starts)
        conditions = Verification.collect(
            condition.verify(context_analyzer, value_context)
            for condition in scenario.conditions
        )
        if not conditions.status.is_succeeded:
            status = Status.SKIPPED
            if conditions.status is Status.FAILURE:
                status = Status.FAILURE

            result = ScenarioResult(label=scenario.label, status=status, conditions=conditions)
            return StaticScenarioTask(result)

        if scenario.ordered:
            submit_cases: Callable = OrderedCasesTask
        else:
            submit_cases = UnorderedCasesTask
        cases = submit_cases(self._executor, self._case_runner, scenario.cases)
        subscenarios = [self.submit(subscenario) for subscenario in scenario.subscenarios]
        return RunningScenarioTask(
            label=scenario.label,
            conditions=conditions,
            cases=cases,
            subscenarios=subscenarios,
        )



from .description import DescriptionCompiler
from .factory import create_description_compiler
from .factory import create_predicate_compiler
from .factory import create_response_description_compiler
from .matcher import MatcherFactoryCompiler
from .predicate import PredicateCompiler
from .response import ResponseDescriptionCompiler, ResponseDescriptionCompiled

__all__ = [
    'DescriptionCompiler',
    'MatcherFactoryCompiler',
    'PredicateCompiler',
    'ResponseDescriptionCompiler',
    'ResponseDescriptionCompiled',
    'create_predicate_compiler',
    'create_description_compiler',
    'create_response_description_compiler',
]



import sys
import types
import uuid
from importlib.util import spec_from_file_location, module_from_spec
from logging import Logger
from typing import Iterable

from pluggy import PluginManager

from preacher.core.logger import default_logger

__all__ = ['load_plugins']


def load_plugins(
    manager: PluginManager,
    plugins: Iterable[str] = (),
    logger: Logger = default_logger,
) -> None:
    """
    Load plugins explicitly.

    Args:
        manager: A plugin manager to load plugins.
        plugins: Plugin paths to load.
        logger: A logger.
    Raises:
        RuntimeException: when a given plugin cannot be loaded as a module.
        Exception: when loading a plugin fails.
    """
    modules = (_load_module(path, logger) for path in plugins)
    for module in modules:
        manager.register(module)


def _load_module(path: str, logger: Logger) -> types.ModuleType:
    name = _unique_name()
    logger.info('Load module file "%s" as module name "%s"', path, name)

    spec = spec_from_file_location(name, path)
    if not spec:
        raise RuntimeError(f'Could not load as a module: {path}')
    module = module_from_spec(spec)
    spec.loader.exec_module(module)  # type: ignore
    sys.modules[name] = module  # To enable sub-processes use this module.

    return module


def _unique_name() -> str:
    return str(uuid.uuid4())



from typing import List, Optional

from preacher.core.request import Response, ExecutionReport
from preacher.core.scenario import ScenarioResult, CaseListener
from preacher.core.status import Status


class Listener(CaseListener):
    """
    Listener interface.
    Default implementations do nothing.
    """

    def on_end(self, status: Status) -> None:
        pass

    def on_scenario(self, result: ScenarioResult) -> None:
        pass


class MergingListener(Listener):

    def __init__(self):
        self._listeners: List[Listener] = []

    def append(self, listener: Listener) -> None:
        self._listeners.append(listener)

    def on_execution(self, execution: ExecutionReport, response: Optional[Response]) -> None:
        for listener in self._listeners:
            listener.on_execution(execution, response)

    def on_scenario(self, result: ScenarioResult) -> None:
        for listener in self._listeners:
            listener.on_scenario(result)

    def on_end(self, status: Status) -> None:
        for listener in self._listeners:
            listener.on_end(status)



from typing import Optional

from pluggy import PluginManager

from preacher.compilation.request import create_request_compiler
from preacher.compilation.verification import create_description_compiler
from preacher.compilation.verification import create_predicate_compiler
from preacher.compilation.verification import create_response_description_compiler
from .case import CaseCompiler
from .scenario import ScenarioCompiler


def create_scenario_compiler(plugin_manager: Optional[PluginManager] = None) -> ScenarioCompiler:
    request = create_request_compiler()

    predicate = create_predicate_compiler(plugin_manager=plugin_manager)
    description = create_description_compiler(predicate=predicate)
    response = create_response_description_compiler(predicate=predicate, description=description)

    case = CaseCompiler(request=request, response=response, description=description)
    return ScenarioCompiler(description=description, case=case)



from typing import Optional

from pluggy import PluginManager

from preacher.compilation.extraction import ExtractionCompiler
from preacher.plugin.manager import get_plugin_manager


def create_extraction_compiler(
    plugin_manager: Optional[PluginManager] = None,
) -> ExtractionCompiler:
    compiler = ExtractionCompiler()

    plugin_manager = plugin_manager or get_plugin_manager()
    plugin_manager.hook.preacher_add_extractions(compiler=compiler)

    return compiler



"""Logging."""

from logging import Logger, getLogger, CRITICAL

__all__ = ['default_logger']


def _create_logger() -> Logger:
    logger = getLogger(__name__)
    logger.setLevel(CRITICAL + 1)  # Do nothing for predefined levels.
    return logger


default_logger: Logger = _create_logger()
"""A default logger for Preacher, which does nothing by default."""



"""
Hamcrest custom matchers.
"""

import operator
from datetime import datetime
from typing import Callable

from hamcrest.core.base_matcher import BaseMatcher
from hamcrest.core.matcher import Matcher
from hamcrest.library.number.ordering_comparison import OrderingComparison

from preacher.core.datetime import DatetimeWithFormat


class _ConvertingMatcher(BaseMatcher):

    def __init__(self, matcher: BaseMatcher, convert: Callable):
        self._matcher = matcher
        self._convert = convert

    def _matches(self, item):
        converted_item = self._convert(item)
        return self._matcher._matches(converted_item)

    def describe_to(self, description):
        self._matcher.describe_to(description)

    def describe_mismatch(self, item, mismatch_description):
        converted_item = self._convert(item)
        mismatch_description.append_text('was ').append_description_of(converted_item)


def before(value: object) -> Matcher:
    """`value` should be a datetime or DateTimeWithFormat."""
    origin = _ensure_datetime(value)
    matcher = OrderingComparison(origin.value, operator.lt, 'before')
    return _ConvertingMatcher(matcher, lambda obj: origin.fmt.parse_datetime(_ensure_str(obj)))


def after(value: object) -> Matcher:
    """`value` should be a datetime or DateTimeWithFormat."""
    origin = _ensure_datetime(value)
    matcher = OrderingComparison(origin.value, operator.gt, 'after')
    return _ConvertingMatcher(matcher, lambda obj: origin.fmt.parse_datetime(_ensure_str(obj)))


def _ensure_str(obj: object) -> str:
    if not isinstance(obj, str):
        raise TypeError(f'Must be a str, but given {type(obj)}: {obj}')
    return obj


def _ensure_datetime(obj: object) -> DatetimeWithFormat:
    if isinstance(obj, DatetimeWithFormat):
        return obj

    if not isinstance(obj, datetime):
        raise TypeError(f'Must be a datetime, but given {type(obj)}: {obj}')
    return DatetimeWithFormat(obj)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from preacher.core.status import Statused, Status, StatusedList
from preacher.core.verification import Verification
from .case_result import CaseResult


@dataclass(frozen=True)
class ScenarioResult(Statused):
    label: Optional[str] = None
    status: Status = Status.SKIPPED
    message: Optional[str] = None
    conditions: Verification = field(default_factory=Verification)
    cases: StatusedList[CaseResult] = field(default_factory=StatusedList)
    subscenarios: StatusedList[ScenarioResult] = field(default_factory=StatusedList)



"""
Value analysis.
"""

import json
from abc import ABC, abstractmethod
from dataclasses import asdict
from typing import Callable, Mapping, Optional, TypeVar

from lxml.etree import _Element as Element, XMLParser, LxmlError, fromstring

from preacher.core.request.response import ResponseBody
from preacher.core.util.functional import recursive_map
from preacher.core.util.serialization import to_serializable
from .error import ExtractionError

T = TypeVar('T')


class Analyzer(ABC):
    """
    Interface to analyze contents.
    """

    @abstractmethod
    def for_text(self, extract: Callable[[str], T]) -> T:
        raise NotImplementedError()

    @abstractmethod
    def for_mapping(self, extract: Callable[[Mapping], T]) -> T:
        raise NotImplementedError()

    @abstractmethod
    def for_etree(self, extract: Callable[[Element], T]) -> T:
        raise NotImplementedError()


class _LazyElementTreeLoader:
    """Loads an element tree from a binary content lazily."""

    def __init__(self, content: bytes):
        self._content = content
        self._is_loaded = False
        self._etree: Optional[Element] = None

    def get(self) -> Element:
        """
        Load an element tree if needed and returns it.

        Returns:
            The loaded element tree.
        Raises:
            ExtractionError: when loading the element tree has failed once.
        """

        if not self._is_loaded:
            try:
                self._etree = fromstring(self._content, parser=XMLParser())
            except LxmlError:
                pass
            self._is_loaded = True

        etree = self._etree
        if etree is None:
            raise ExtractionError('Not an XML content')
        return etree


class _LazyJsonLoader:
    """Loads a JSON value from a text lazily."""

    def __init__(self, text: str):
        self._text = text
        self._is_loaded = False
        self._is_valid_json = False
        self._value: object = None

    def get(self) -> object:
        """
        Load a JSON value if needed and returns it.

        Returns:
            The loaded JSON value.
        Raises:
            ExtractionError: when loading the JSON value has failed once.
        """

        if not self._is_loaded:
            try:
                self._value = json.loads(self._text)
                self._is_valid_json = True
            except json.JSONDecodeError:
                pass
            self._is_loaded = True

        if not self._is_valid_json:
            raise ExtractionError('Not a JSON content')
        return self._value


class ResponseBodyAnalyzer(Analyzer):

    def __init__(self, body: ResponseBody):
        self._body = body
        self._etree_loader = _LazyElementTreeLoader(body.content)
        self._json_loader = _LazyJsonLoader(body.text)

    def for_text(self, extract: Callable[[str], T]) -> T:
        return extract(self._body.text)

    def for_mapping(self, extract: Callable[[Mapping], T]) -> T:
        json_value = self._json_loader.get()
        if not isinstance(json_value, Mapping):
            raise ExtractionError(f'Expected a dictionary, but given {type(json_value)}')
        return extract(json_value)

    def for_etree(self, extract: Callable[[Element], T]) -> T:
        return extract(self._etree_loader.get())


class MappingAnalyzer(Analyzer):

    def __init__(self, value: Mapping):
        self._value = value

    def for_text(self, extract: Callable[[str], T]) -> T:
        serializable = recursive_map(to_serializable, self._value)
        return extract(json.dumps(serializable, separators=(',', ':')))

    def for_mapping(self, extract: Callable[[Mapping], T]) -> T:
        return extract(self._value)

    def for_etree(self, extract: Callable[[Element], T]) -> T:
        raise ExtractionError('Not an XML content')


def analyze_data_obj(obj) -> Analyzer:
    return MappingAnalyzer(asdict(obj))



"""CLI Application implementation."""

import sys
from concurrent.futures import Executor, ProcessPoolExecutor
from itertools import chain
from logging import DEBUG, INFO, WARNING, ERROR
from logging import Logger, StreamHandler, getLogger
from typing import Callable, Iterable, Iterator, Optional, Sequence

from preacher.compilation.argument import Arguments
from preacher.compilation.scenario import create_scenario_compiler
from preacher.compilation.yaml import load_all, load_all_from_path
from preacher.core.logger import default_logger
from preacher.core.request import Requester
from preacher.core.scenario import ScenarioRunner, CaseRunner
from preacher.core.scheduling import ScenarioScheduler, Listener, MergingListener
from preacher.core.status import Status
from preacher.core.unit import UnitRunner
from preacher.plugin.loader import load_plugins
from preacher.plugin.manager import get_plugin_manager
from preacher.presentation.listener import LoggingReportingListener, HtmlReportingListener
from .logging import ColoredFormatter

__all__ = ['app', 'create_system_logger', 'create_listener', 'create_scheduler', 'load_objs']

_REPORT_LOGGER_NAME = 'preacher.cli.report.logging'


def app(
    paths: Sequence[str] = (),
    base_url: str = '',
    arguments: Optional[Arguments] = None,
    level: Status = Status.SUCCESS,
    report_dir: Optional[str] = None,
    delay: float = 0.1,
    retry: int = 0,
    timeout: Optional[float] = None,
    concurrency: int = 1,
    executor_factory: Callable[[int], Executor] = ProcessPoolExecutor,
    plugins: Iterable[str] = (),
    verbosity: int = 0,
) -> int:
    """
    Preacher CLI application.

    Returns:
        the exit code.
    """

    # Fill default.
    arguments = arguments or {}

    logger = create_system_logger(verbosity)

    logger.debug(
        'Running condition\n'
        '  Paths: %s\n'
        '  Arguments: %s\n'
        '  Base URL: %s\n'
        '  Logging report level: %s\n'
        '  Reporting directory path: %s\n'
        '  Max retry count: %d\n'
        '  Delay between attempts in seconds: %s\n'
        '  Timeout in seconds: %s\n'
        '  Concurrency: %s\n'
        '  Executor: %s\n'
        '  Verbosity: %d',
        paths,
        arguments,
        base_url,
        level,
        report_dir,
        retry,
        delay,
        timeout,
        concurrency,
        executor_factory,
        verbosity
    )

    plugin_manager = get_plugin_manager()
    try:
        load_plugins(plugin_manager, plugins, logger)
        compiler = create_scenario_compiler(plugin_manager=plugin_manager)
    except Exception as error:
        logger.exception(error)
        return 3

    objs = load_objs(paths, logger)
    scenario_groups = (compiler.compile_flattening(obj, arguments=arguments) for obj in objs)
    scenarios = chain.from_iterable(scenario_groups)

    listener = create_listener(level, report_dir)
    try:
        logger.info('Start running scenarios.')
        with executor_factory(concurrency) as executor:
            scheduler = create_scheduler(
                executor=executor,
                listener=listener,
                base_url=base_url,
                timeout=timeout,
                retry=retry,
                delay=delay,
            )
            status = scheduler.run(scenarios)
    except Exception as error:
        logger.exception(error)
        return 3
    finally:
        logger.info('End running scenarios.')

    if not status.is_succeeded:
        return 1

    return 0


def create_system_logger(verbosity: int) -> Logger:
    level = _verbosity_to_logging_level(verbosity)
    handler = StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(ColoredFormatter(fmt='[%(levelname)s] %(message)s'))
    logger = getLogger(__name__)
    logger.setLevel(level)
    logger.addHandler(handler)
    return logger


def _verbosity_to_logging_level(verbosity: int) -> int:
    if verbosity > 1:
        return DEBUG
    if verbosity > 0:
        return INFO
    return WARNING


def load_objs(paths: Sequence[str], logger: Logger = default_logger) -> Iterator[object]:
    if not paths:
        logger.info('No scenario file is given. Load scenarios from stdin.')
        return load_all(sys.stdin)
    return chain.from_iterable(load_all_from_path(_hook_loading(path, logger)) for path in paths)


def _hook_loading(path: str, logger: Logger) -> str:
    logger.debug('Load: %s', path)
    return path


def create_listener(level: Status, report_dir: Optional[str]) -> Listener:
    merging = MergingListener()

    logging_level = _status_to_logging_level(level)
    handler = StreamHandler(sys.stdout)
    handler.setLevel(logging_level)
    handler.setFormatter(ColoredFormatter())
    logger = getLogger(_REPORT_LOGGER_NAME)
    logger.setLevel(logging_level)
    logger.addHandler(handler)
    merging.append(LoggingReportingListener.from_logger(logger))

    if report_dir:
        merging.append(HtmlReportingListener.from_path(report_dir))

    return merging


def _status_to_logging_level(level: Status) -> int:
    if level is Status.SKIPPED:
        return DEBUG
    if level is Status.SUCCESS:
        return INFO
    if level is Status.UNSTABLE:
        return WARNING
    return ERROR


def create_scheduler(
    executor: Executor,
    listener: Listener,
    base_url: str,
    timeout: Optional[float],
    retry: int,
    delay: float,
) -> ScenarioScheduler:
    requester = Requester(base_url=base_url, timeout=timeout)
    unit_runner = UnitRunner(requester=requester, retry=retry, delay=delay)
    case_runner = CaseRunner(unit_runner=unit_runner, listener=listener)
    runner = ScenarioRunner(executor=executor, case_runner=case_runner)
    return ScenarioScheduler(runner=runner, listener=listener)



"""
YAML Error handling.
"""

from contextlib import contextmanager
from typing import Optional, Iterator

from yaml import Mark, Node


class YamlError(Exception):

    def __init__(
        self,
        message: Optional[str] = None,
        mark: Optional[Mark] = None,
        cause: Optional[Exception] = None,
    ):
        self._message = message
        self._mark = mark
        self._cause = cause

    def __str__(self) -> str:
        lines = []

        if self._message is not None:
            lines.append(self._message)
        if self._cause:
            lines.append(str(self._cause))
        if self._mark:
            lines.append(str(self._mark))
        return '\n'.join(lines)


@contextmanager
def on_node(node: Node) -> Iterator:
    try:
        yield
    except Exception as error:
        raise YamlError(mark=node.start_mark, cause=error)



"""Request."""

from .factory import create_request_compiler
from .request import RequestCompiler, RequestCompiled
from .request_body import RequestBodyCompiler, RequestBodyCompiled
from .url_param import compile_url_params

__all__ = [
    'RequestCompiler',
    'RequestCompiled',
    'RequestBodyCompiler',
    'RequestBodyCompiled',
    'compile_url_params',
    'create_request_compiler',
]



from abc import ABC, abstractmethod
from concurrent.futures import Executor
from typing import Iterable

from requests import Session

from preacher.core.scenario.case import Case
from preacher.core.scenario.case_result import CaseResult
from preacher.core.scenario.case_runner import CaseRunner
from preacher.core.status import StatusedList


class CasesTask(ABC):

    @abstractmethod
    def result(self) -> StatusedList[CaseResult]:
        raise NotImplementedError()


def _run_cases_in_order(
    case_runner: CaseRunner,
    cases: Iterable[Case],
    *args,
    **kwargs,
) -> StatusedList[CaseResult]:
    if not kwargs.get('session'):
        with Session() as session:
            kwargs['session'] = session
            return _run_cases_in_order(case_runner, cases, *args, **kwargs)

    return StatusedList.collect(case_runner.run(case, *args, **kwargs) for case in cases)


class OrderedCasesTask(CasesTask):

    def __init__(
        self,
        executor: Executor,
        case_runner: CaseRunner,
        cases: Iterable[Case],
        *args,
        **kwargs,
    ):
        self._future = executor.submit(
            _run_cases_in_order,
            case_runner,
            cases,
            *args,
            **kwargs,
        )

    def result(self) -> StatusedList[CaseResult]:
        return self._future.result()


class UnorderedCasesTask(CasesTask):

    def __init__(
        self,
        executor: Executor,
        case_runner: CaseRunner,
        cases: Iterable[Case],
        *args,
        **kwargs,
    ):
        self._futures = [
            executor.submit(case_runner.run, case, *args, **kwargs)
            for case in cases
        ]

    def result(self) -> StatusedList[CaseResult]:
        return StatusedList.collect(f.result() for f in self._futures)



"""
Datetime utilities for Preacher core.
"""

import time
from abc import ABC
from datetime import datetime, timedelta, timezone
from typing import Optional

from dateutil.parser import isoparse


class DatetimeFormat(ABC):

    def format_datetime(self, value: datetime) -> str:
        raise NotImplementedError()

    def parse_datetime(self, value: str) -> datetime:
        raise NotImplementedError()


class Iso8601Format(DatetimeFormat):

    def format_datetime(self, value: datetime) -> str:
        return value.isoformat()

    def parse_datetime(self, value: str) -> datetime:
        try:
            return isoparse(value)
        except ValueError as error:
            raise ValueError(f'An invalid ISO 8601 format: {value}', error)


class StrftimeFormat(DatetimeFormat):

    def __init__(self, format_string: str):
        self._format_string = format_string

    def format_datetime(self, value: datetime) -> str:
        return value.strftime(self._format_string)

    def parse_datetime(self, value: str) -> datetime:
        return datetime.strptime(value, self._format_string)


ISO8601 = Iso8601Format()


class DatetimeWithFormat:

    def __init__(
        self,
        value: datetime,
        fmt: Optional[DatetimeFormat] = None,
    ):
        self._value = value
        self._fmt = fmt or ISO8601

    @property
    def value(self) -> datetime:
        return self._value

    @property
    def fmt(self) -> DatetimeFormat:
        return self._fmt

    @property
    def formatted(self) -> str:
        return self._fmt.format_datetime(self._value)


def now() -> datetime:
    return datetime.now(_system_timezone())


def _system_timezone() -> timezone:
    localtime = time.localtime()
    return timezone(
        offset=timedelta(seconds=localtime.tm_gmtoff),
        name=localtime.tm_zone,
    )



import logging

from colorama import Fore, Style, init

init()


def _default(message: str) -> str:
    return message


def _info(message: str) -> str:
    return f'{Fore.GREEN}{message}{Style.RESET_ALL}'


def _warning(message: str) -> str:
    return f'{Fore.YELLOW}{message}{Style.RESET_ALL}'


def _error(message: str) -> str:
    return f'{Fore.RED}{message}{Style.RESET_ALL}'


def _critical(message: str) -> str:
    return f'{Fore.RED}{Style.BRIGHT}{message}{Style.RESET_ALL}'


_STYLE_FUNC_MAP = {
    logging.INFO: _info,
    logging.WARNING: _warning,
    logging.ERROR: _error,
    logging.CRITICAL: _critical,
}


class ColoredFormatter(logging.Formatter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        style_func = _STYLE_FUNC_MAP.get(record.levelno, _default)
        return style_func(super().format(record))



from dataclasses import dataclass, field
from typing import Optional

from .argument import Arguments
from .error import on_key
from .util.type import ensure_optional_str, ensure_mapping

_KEY_LABEL = 'label'
_KEY_ARGUMENTS = 'args'


@dataclass(frozen=True)
class Parameter:
    label: Optional[str] = None
    arguments: Arguments = field(default_factory=dict)


def compile_parameter(obj: object) -> Parameter:
    """
    Compile a parameter.

    Args:
        obj: A compiled object.
    Returns:
        A parameter as the result of compilation.
    Raises:
        CompilationError: when the compilation fails.
    """
    obj = ensure_mapping(obj)

    with on_key(_KEY_LABEL):
        label = ensure_optional_str(obj.get(_KEY_LABEL))
    with on_key(_KEY_ARGUMENTS):
        arguments = ensure_mapping(obj.get(_KEY_ARGUMENTS, {}))
    return Parameter(label=label, arguments=arguments)



"""Request compilation."""

from .request import Request, Method
from .requester import Requester, ExecutionReport, PreparedRequest
from .request_body import RequestBody, UrlencodedRequestBody, JsonRequestBody
from .response import Response, ResponseBody
from .url_param import UrlParams, UrlParam, UrlParamValue

__all__ = [
    'Request',
    'Method',
    'RequestBody',
    'UrlencodedRequestBody',
    'JsonRequestBody',
    'Response',
    'ResponseBody',
    'UrlParams',
    'UrlParam',
    'UrlParamValue',
    'Requester',
    'ExecutionReport',
    'PreparedRequest',
]



from yaml import BaseLoader, Node

from preacher.compilation.argument import Argument
from preacher.compilation.util.type import ensure_str
from .error import on_node


def construct_argument(loader: BaseLoader, node: Node) -> Argument:
    obj = loader.construct_scalar(node)
    with on_node(node):
        key = ensure_str(obj)
    return Argument(key)



"""Response."""

from abc import ABC, abstractmethod
from typing import Mapping


class ResponseBody(ABC):

    @property
    @abstractmethod
    def text(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def content(self) -> bytes:
        raise NotImplementedError()


class Response(ABC):

    @property
    @abstractmethod
    def id(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def elapsed(self) -> float:
        raise NotImplementedError()

    @property
    @abstractmethod
    def status_code(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def headers(self) -> Mapping[str, str]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def body(self) -> ResponseBody:
        raise NotImplementedError()



"""Status."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass, field
from enum import Enum
from functools import reduce
from typing import Generic, List, TypeVar


class Status(Enum):

    # Numbers stand for the priorities for merging.
    SKIPPED = 0
    SUCCESS = 1
    UNSTABLE = 2
    FAILURE = 3

    @property
    def is_succeeded(self):
        return self.value <= Status.SUCCESS.value

    def merge(self, other: Status):
        return max(self, other, key=lambda status: status.value)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return str(self)

    def __bool__(self) -> bool:
        return self.is_succeeded


def merge_statuses(statuses: Iterable[Status]) -> Status:
    return reduce(lambda lhs, rhs: lhs.merge(rhs), statuses, Status.SKIPPED)


class Statused(ABC):

    @property
    @abstractmethod
    def status(self) -> Status:
        raise NotImplementedError()


StatusedType = TypeVar('StatusedType', bound=Statused)


@dataclass(frozen=True)
class StatusedList(Generic[StatusedType], Statused):
    items: List[StatusedType] = field(default_factory=list)

    @property
    def status(self) -> Status:  # HACK: should be cached
        return merge_statuses(item.status for item in self.items)

    @staticmethod
    def collect(
        iterable: Iterable[StatusedType],
    ) -> StatusedList[StatusedType]:
        return StatusedList(list(iterable))



"""Typing utilities."""

from functools import partial
from typing import Callable, Type, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def require_type(tp: Type[T], func: Callable[[T], U]) -> Callable[[object], U]:
    """
    Returns a function
    that asserts its argument should be an instance of `tp`:
    when given an argument that is not an instance of `tp`,
    then throws `InterpretationError`.
    """
    return partial(_require_type, tp, func)


def _require_type(tp: Type[T], func: Callable[[T], U], value: object) -> U:
    if not isinstance(value, tp):
        raise TypeError(f"Argument 1 must be a {tp}")
    return func(value)



from abc import ABC, abstractmethod
from typing import Optional, List

from preacher.core.status import StatusedList, merge_statuses
from preacher.core.verification import Verification
from .scenario_result import ScenarioResult
from .util.concurrency import CasesTask


class ScenarioTask(ABC):

    @abstractmethod
    def result(self) -> ScenarioResult:
        raise NotImplementedError()


class StaticScenarioTask(ScenarioTask):

    def __init__(self, result: ScenarioResult):
        self._result = result

    def result(self) -> ScenarioResult:
        return self._result


class RunningScenarioTask(ScenarioTask):

    def __init__(
        self,
        label: Optional[str],
        conditions: Verification,
        cases: CasesTask,
        subscenarios: List[ScenarioTask],
    ):
        self._label = label
        self._conditions = conditions
        self._cases = cases
        self._subscenarios = subscenarios

    def result(self) -> ScenarioResult:
        cases = self._cases.result()
        subscenarios = StatusedList.collect(s.result() for s in self._subscenarios)
        return ScenarioResult(
            label=self._label,
            status=merge_statuses([cases.status, subscenarios.status]),
            conditions=self._conditions,
            cases=cases,
            subscenarios=subscenarios,
        )



"""URL Parameters."""

from datetime import date, datetime
from typing import Union, List, Mapping, Optional

from preacher.core.datetime import DatetimeWithFormat
from preacher.core.value import Value, ValueContext

UrlParamValue = Union[
    None,
    bool,
    int,
    float,
    str,
    date,
    datetime,
    DatetimeWithFormat,
    Value[None],
    Value[bool],
    Value[int],
    Value[float],
    Value[str],
    Value[date],
    Value[datetime],
    Value[DatetimeWithFormat],
]
UrlParam = Union[UrlParamValue, List[UrlParamValue]]
UrlParams = Union[str, Mapping[str, UrlParam]]
ResolvedUrlParamValue = Optional[str]
ResolvedUrlParam = Union[ResolvedUrlParamValue, List[ResolvedUrlParamValue]]
ResolvedUrlParams = Union[str, Mapping[str, ResolvedUrlParam]]


def resolve_url_param_value(
    value: UrlParamValue,
    context: Optional[ValueContext] = None,
) -> Optional[str]:
    if isinstance(value, Value):
        value = value.resolve(context)

    if value is None:
        return None
    if isinstance(value, bool):
        return 'true' if value else 'false'
    if isinstance(value, DatetimeWithFormat):
        return value.formatted
    if isinstance(value, date):
        return value.isoformat()
    return str(value)


def resolve_url_param(
    param: UrlParam,
    context: Optional[ValueContext] = None,
) -> ResolvedUrlParam:
    if isinstance(param, list):
        return [resolve_url_param_value(value, context) for value in param]
    return resolve_url_param_value(param)


def resolve_url_params(
    params: UrlParams,
    context: Optional[ValueContext] = None,
) -> ResolvedUrlParams:
    if isinstance(params, str):
        return params
    return {
        key: resolve_url_param(param, context)
        for (key, param) in params.items()
    }



from typing import Optional

from pluggy import PluginManager

from preacher.compilation.extraction import create_extraction_compiler
from preacher.plugin.manager import get_plugin_manager
from .description import DescriptionCompiler
from .matcher import MatcherFactoryCompiler
from .predicate import PredicateCompiler
from .response import ResponseDescriptionCompiler


def create_matcher_factory_compiler(
    plugin_manager: Optional[PluginManager] = None,
) -> MatcherFactoryCompiler:
    compiler = MatcherFactoryCompiler()

    plugin_manager = plugin_manager or get_plugin_manager()
    plugin_manager.hook.preacher_add_matchers(compiler=compiler)

    return compiler


def create_predicate_compiler(
    plugin_manager: Optional[PluginManager] = None,
) -> PredicateCompiler:
    matcher_factory = create_matcher_factory_compiler(plugin_manager=plugin_manager)
    return PredicateCompiler(matcher_factory)


def create_description_compiler(
    predicate: Optional[PredicateCompiler] = None,
    plugin_manager: Optional[PluginManager] = None,
) -> DescriptionCompiler:
    predicate = predicate or create_predicate_compiler(plugin_manager=plugin_manager)
    extraction = create_extraction_compiler(plugin_manager=plugin_manager)

    return DescriptionCompiler(extraction=extraction, predicate=predicate)


def create_response_description_compiler(
    predicate: Optional[PredicateCompiler] = None,
    description: Optional[DescriptionCompiler] = None,
    plugin_manager: Optional[PluginManager] = None,
) -> ResponseDescriptionCompiler:
    predicate = predicate or create_predicate_compiler(plugin_manager=plugin_manager)
    description = description or create_description_compiler()

    return ResponseDescriptionCompiler(predicate=predicate, description=description)



"""Predicate compilation."""

from preacher.core.verification.matcher import MatcherWrappingPredicate
from preacher.core.verification.predicate import Predicate
from .matcher import MatcherFactoryCompiler


class PredicateCompiler:

    def __init__(self, matcher_factory: MatcherFactoryCompiler):
        self._matcher_factory = matcher_factory

    def compile(self, obj: object) -> Predicate:
        factory = self._matcher_factory.compile(obj)
        return MatcherWrappingPredicate(factory)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['preacher',
 'preacher.app',
 'preacher.app.cli',
 'preacher.compilation',
 'preacher.compilation.extraction',
 'preacher.compilation.request',
 'preacher.compilation.scenario',
 'preacher.compilation.util',
 'preacher.compilation.verification',
 'preacher.compilation.yaml',
 'preacher.core',
 'preacher.core.extraction',
 'preacher.core.extraction.impl',
 'preacher.core.request',
 'preacher.core.scenario',
 'preacher.core.scenario.util',
 'preacher.core.scheduling',
 'preacher.core.unit',
 'preacher.core.util',
 'preacher.core.verification',
 'preacher.plugin',
 'preacher.presentation',
 'preacher.presentation.listener']

package_data = \
{'': ['*'],
 'preacher': ['resources/report/html/*', 'resources/report/html/macros/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'colorama>=0.4.1,<0.5.0',
 'jinja2>=2.11,<3.0',
 'jq>=1.1.2,<2.0.0',
 'lxml>=4.4,<5.0',
 'pluggy>=0.13.1,<0.14.0',
 'pyhamcrest>=2.0,<3.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.21,<3.0']

entry_points = \
{'console_scripts': ['preacher-cli = preacher.app.cli.main:main']}

setup_kwargs = {
    'name': 'preacher',
    'version': '0.15.3',
    'description': 'Web API Verification without Coding.',
    'long_description': '# Preacher: Web API Verification without Coding\n\n[![PyPI version](https://badge.fury.io/py/preacher.svg)][PyPI]\n[![Documentation Status](https://readthedocs.org/projects/preacher/badge/?version=latest)][Read the Docs]\n[![CircleCI](https://circleci.com/gh/ymoch/preacher.svg?style=svg)][Circle CI]\n[![Codecov](https://codecov.io/gh/ymoch/preacher/branch/master/graph/badge.svg)][Codecov]\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/ymoch/preacher.svg?logo=lgtm&logoWidth=18)][LGTM]\n\nPreacher verifies API servers,\nwhich requests to the servers and verify the responses along to given scenarios.\n\nTest scenarios are written only in [YAML][] declaratively, without coding.\nIn spite of that, Preacher can validate your web API flexibly,\nwhich enables you to test using real (neither mocks nor sandboxes) backends.\n\n- Responses are analyzed [jq][] or [XPath][] queries\n- Validation rules are based on [Hamcrest][] (implemented by [PyHamcrest][]).\n\nThe full documentation is available at [preacher.readthedocs.io][Read the Docs].\n\n## Targets\n\n- Flexible validation to test with real backends: neither mocks nor sandboxes.\n  - Matcher-based validation.\n- CI Friendly to automate easily.\n  - A CLI application and YAML-based scenarios.\n\n## Usage\n\nFirst, install Preacher.\n\nThe most basic way to install Preacher is using `pip`. Supports only Python 3.7+.\n\n```sh\n$ pip install preacher\n$ preacher-cli --version\n```\n\nInstead of `pip`, Docker images are also available on\n[Docker Hub](https://hub.docker.com/r/ymoch/preacher)\nas `ymoch/preacher`.\nBy default, the container working directory is `/work`,\nand the host directory may be mounted here.\n\n```sh\n$ docker pull ymock/preacher\n$ docker run -v $PWD:/work ymoch/preacher preacher-cli --version\n```\n\nSecond, write your own scenario.\n\n```yaml\n# scenario.yml\nlabel: An example of a scenario\ncases:\n  - label: An example of a case\n    request: /path/to/foo\n    response:\n      status_code: 200\n      body:\n        - describe: .foo\n          should:\n            equal: bar\n```\n\nThen, run ``preacher-cli`` command.\n\n```sh\n$ preacher-cli -u http://your.domain.com/base scenario.yml\n```\n\nFor more information such as grammer of scenarios,\nsee [the full documentation][Read the Docs].\n\n## License\n\n[![MIT License](https://img.shields.io/badge/License-MIT-brightgreen.svg)][MIT License]\n\nCopyright (c) 2019 Yu MOCHIZUKI\n\n\n[YAML]: https://yaml.org/\n[jq]: https://stedolan.github.io/jq/\n[XPath]: https://www.w3.org/TR/xpath/all/\n[Hamcrest]: http://hamcrest.org/\n[PyHamcrest]: https://pyhamcrest.readthedocs.io/\n[MIT License]: https://opensource.org/licenses/MIT\n\n[Read the Docs]: https://preacher.readthedocs.io/\n[PyPI]: https://badge.fury.io/py/preacher\n[Circle CI]: https://circleci.com/gh/ymoch/preacher\n[Codecov]: https://codecov.io/gh/ymoch/preacher\n[LGTM]: https://lgtm.com/projects/g/ymoch/preacher/context:python\n',
    'author': 'Yu Mochizuki',
    'author_email': 'ymoch.dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://preacher.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""YAML loaders."""

import glob
import os
import re
from contextlib import contextmanager
from typing import Iterator, TextIO

from yaml import Node, BaseLoader, MarkedYAMLError
from yaml import load as _load, load_all as _load_all
from yaml.composer import Composer
from yaml.constructor import SafeConstructor
from yaml.parser import Parser
from yaml.reader import Reader
from yaml.resolver import Resolver
from yaml.scanner import Scanner

from preacher.compilation.util.type import ensure_str
from .argument import construct_argument
from .datetime import construct_relative_datetime
from .error import YamlError, on_node

_WILDCARDS_REGEX = re.compile(r'^.*(\*|\?|\[!?.+\]).*$')


class Loader:

    def __init__(self):
        self._origin = '.'

        class _Ctor(SafeConstructor):
            pass

        _Ctor.add_constructor('!include', self._include)
        _Ctor.add_constructor('!argument', construct_argument)
        _Ctor.add_constructor(
            '!relative_datetime',
            construct_relative_datetime,
        )

        class _Loader(Reader, Scanner, Parser, Composer, _Ctor, Resolver):
            def __init__(self, stream):
                Reader.__init__(self, stream)
                Scanner.__init__(self)
                Parser.__init__(self)
                Composer.__init__(self)
                SafeConstructor.__init__(self)

        self._Loader = _Loader

    def load(self, stream: TextIO, origin: str = '.') -> object:
        try:
            with self._on_origin(origin):
                return _load(stream, self._Loader)
        except MarkedYAMLError as error:
            raise YamlError(cause=error)

    def load_from_path(self, path: str) -> object:
        origin = os.path.dirname(path)
        try:
            with open(path) as stream:
                return self.load(stream, origin)
        except FileNotFoundError as error:
            raise YamlError(cause=error)

    def load_all(self, stream: TextIO, origin: str = '.') -> Iterator:
        try:
            with self._on_origin(origin):
                yield from _load_all(stream, self._Loader)
        except MarkedYAMLError as error:
            raise YamlError(cause=error)

    def load_all_from_path(self, path: str) -> Iterator:
        origin = os.path.dirname(path)
        try:
            with open(path) as stream:
                yield from self.load_all(stream, origin)
        except FileNotFoundError as error:
            raise YamlError(cause=error)

    def _include(self, loader: BaseLoader, node: Node) -> object:
        obj = loader.construct_scalar(node)

        with on_node(node):
            base = ensure_str(obj)
            path = os.path.join(self._origin, base)
            if _WILDCARDS_REGEX.match(path):
                paths = glob.iglob(path, recursive=True)
                return [self.load_from_path(p) for p in paths]
            return self.load_from_path(path)

    @contextmanager
    def _on_origin(self, origin: str) -> Iterator:
        original = self._origin
        self._origin = origin
        try:
            yield
        finally:
            self._origin = original


def load(stream: TextIO, origin: str = '.') -> object:
    return Loader().load(stream, origin)


def load_from_path(path: str) -> object:
    return Loader().load_from_path(path)


def load_all(stream: TextIO, origin: str = '.') -> Iterator:
    return Loader().load_all(stream, origin)


def load_all_from_path(path: str) -> Iterator:
    return Loader().load_all_from_path(path)



from .case import CaseCompiler, CaseCompiled
from .factory import create_scenario_compiler
from .scenario import ScenarioCompiler

__all__ = [
    'CaseCompiler',
    'CaseCompiled',
    'ScenarioCompiler',
    'create_scenario_compiler',
]



"""CLI Options."""

import logging
import re
import shlex
from concurrent.futures import (
    Executor,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
)
from enum import IntEnum
from typing import Callable, Iterable, Mapping, Optional, Tuple, Union

from click import (
    Choice,
    Context,
    BadParameter,
    Parameter,
    ParamType,
    Option,
    STRING,
)
from yaml import safe_load
from yaml.error import MarkedYAMLError

from preacher.compilation.argument import Arguments
from preacher.core.status import Status


class Level(IntEnum):
    SKIPPED = logging.DEBUG
    SUCCESS = logging.INFO
    UNSTABLE = logging.WARN
    FAILURE = logging.ERROR

    def __str__(self):
        return self.name.lower()


_LEVEL_MAP: Mapping[str, Level] = {str(level): level for level in Level}
_CONCURRENT_EXECUTOR_FACTORY_MAP: Mapping[str, Callable[[int], Executor]] = {
    'process': ProcessPoolExecutor,
    'thread': ThreadPoolExecutor,
}


class ArgumentType(ParamType):

    _base = STRING
    name = _base.name

    def convert(self, value, param, ctx):
        exp = self._base.convert(value, param, ctx)
        return _parse_argument(exp)

    def split_envvar_value(self, rv):
        return shlex.split(rv)


class LevelType(ParamType):

    _choice = Choice(
        tuple(item.name.lower() for item in Status),
        case_sensitive=False,
    )
    name = _choice.name

    def get_metavar(self, param):
        return self._choice.get_metavar(param)

    def get_missing_message(self, param):
        return self._choice.get_missing_message(param)

    def convert(self, value, param, ctx):
        key = self._choice.convert(value, param, ctx).lower()
        return next(item for item in Status if item.name.lower() == key)


class ExecutorFactoryType(ParamType):

    _choice = Choice(
        tuple(_CONCURRENT_EXECUTOR_FACTORY_MAP.keys()),
        case_sensitive=False,
    )
    name = _choice.name

    def get_metavar(self, param):
        return self._choice.get_metavar(param)

    def get_missing_message(self, param):
        return self._choice.get_missing_message(param)

    def convert(self, value, param, ctx):
        key = self._choice.convert(value, param, ctx)
        return _CONCURRENT_EXECUTOR_FACTORY_MAP[key.lower()]


def pairs_callback(
    _context: Context,
    _option_or_parameter: Union[Option, Parameter],
    value: Iterable[Tuple[str, object]],
) -> Arguments:
    return dict(value)


def positive_float_callback(
    _context: Context,
    _option_or_parameter: Union[Option, Parameter],
    value: Optional[float],
) -> Optional[float]:
    if value is None:
        return value

    if value <= 0.0:
        raise BadParameter(f'must be positive, given {value}')
    return value


def _parse_argument(value: str) -> Tuple[str, object]:
    match = re.match(r'^([^=]+)=(.*)$', value)
    if not match:
        raise BadParameter(f'Invalid format argument: {value}')

    key = match.group(1)
    try:
        value = safe_load(match.group(2))
    except MarkedYAMLError as error:
        raise BadParameter(f'Invalid YAML format: {value}\n{error}')

    return key, value



import os
from typing import Iterable

import jinja2

from preacher.core.request import Response, ExecutionReport
from preacher.core.scenario import ScenarioResult


class HtmlReporter:

    def __init__(self, path):
        self._path = path
        self._responses_path = os.path.join(self._path, 'responses')

        loader = jinja2.PackageLoader('preacher', 'resources/report/html')
        self._env = jinja2.Environment(loader=loader, autoescape=True)

        self._initialize()

    def _initialize(self) -> None:
        os.makedirs(self._path, exist_ok=True)
        os.makedirs(self._responses_path, exist_ok=True)

    def export_response(self, execution: ExecutionReport, response: Response) -> None:
        name = f'{response.id}.html'
        path = os.path.join(self._responses_path, name)

        template = self._env.get_template('response-view.html')
        with open(path, 'w') as f:
            template.stream(execution=execution, response=response).dump(f)

    def export_results(self, results: Iterable[ScenarioResult]) -> None:
        html_path = os.path.join(self._path, 'index.html')

        template = self._env.get_template('index.html')
        with open(html_path, 'w') as f:
            template.stream(scenarios=results).dump(f)



from typing import Optional, Callable, Any, List

from lxml.etree import _Element as Element, XPathEvalError

from preacher.core.extraction import Analyzer, ExtractionError
from preacher.core.extraction.extraction import Extractor
from preacher.core.util.functional import identity, apply_if_not_none


class XPathExtractor(Extractor):

    def __init__(
        self,
        query: str,
        multiple: bool = False,
        cast: Optional[Callable[[object], Any]] = None,
    ):
        self._query = query
        self._multiple = multiple
        self._cast = cast or identity

    def extract(self, analyzer: Analyzer) -> object:
        elements = analyzer.for_etree(self._extract)
        if not elements:
            return None

        values = (apply_if_not_none(self._cast, self._convert(element)) for element in elements)
        if self._multiple:
            return list(values)
        else:
            return next(values, None)

    @staticmethod
    def _convert(elem: object) -> str:
        if isinstance(elem, Element):
            return elem.text
        return str(elem)

    def _extract(self, elem: Element) -> List[Element]:
        try:
            return elem.xpath(self._query)
        except XPathEvalError:
            raise ExtractionError(f'Invalid XPath: {self._query}')



from typing import Optional, Callable, Any

from preacher.core.extraction import Analyzer
from preacher.core.extraction.extraction import Extractor
from preacher.core.util.functional import identity, apply_if_not_none


class KeyExtractor(Extractor):

    def __init__(
        self,
        key: str,
        multiple: bool = False,
        cast: Optional[Callable[[object], Any]] = None,
    ):
        self._key = key
        self._multiple = multiple
        self._cast = cast or identity

    def extract(self, analyzer: Analyzer) -> object:
        value = apply_if_not_none(self._cast, analyzer.for_mapping(lambda v: v.get(self._key)))
        return [value] if self._multiple else value



"""
Test cases, which execute a given request and verify its response
along the given descriptions.
"""

from typing import Optional, List

from preacher.core.request import Request
from preacher.core.verification import Description
from preacher.core.verification import ResponseDescription


class Case:
    """
    Test cases, which execute a given request and verify its response
    along the given descriptions.
    """

    def __init__(
        self,
        label: Optional[str] = None,
        enabled: bool = True,
        conditions: Optional[List[Description]] = None,
        request: Optional[Request] = None,
        response: Optional[ResponseDescription] = None,
    ):
        self._label = label
        self._enabled = enabled
        self._conditions = conditions or []
        self._request = request or Request()
        self._response = response or ResponseDescription()

    @property
    def label(self) -> Optional[str]:
        return self._label

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def conditions(self) -> List[Description]:
        return self._conditions

    @property
    def request(self) -> Request:
        return self._request

    @property
    def response(self) -> ResponseDescription:
        return self._response



from preacher.core.scheduling.listener import Listener, MergingListener
from preacher.core.scheduling.scenario_scheduler import ScenarioScheduler

__all__ = [
    'ScenarioScheduler',
    'Listener',
    'MergingListener',
]



"""Preacher CLI."""

import sys
from concurrent.futures import Executor
from typing import Callable, Iterable, Optional, Sequence

from click import FloatRange
from click import IntRange
from click import Path
from click import argument
from click import command
from click import help_option
from click import option
from click import version_option

from preacher import __version__ as _version
from preacher.compilation.argument import Arguments
from preacher.core.status import Status
from .app import app
from .option import ArgumentType
from .option import ExecutorFactoryType
from .option import LevelType
from .option import pairs_callback
from .option import positive_float_callback

_ENV_PREFIX = 'PREACHER_CLI_'
_ENV_BASE_URL = f'{_ENV_PREFIX}BASE_URL'
_ENV_ARGUMENT = f'{_ENV_PREFIX}ARGUMENT'
_ENV_LEVEL = f'{_ENV_PREFIX}LEVEL'
_ENV_RETRY = f'{_ENV_PREFIX}RETRY'
_ENV_DELAY = f'{_ENV_PREFIX}DELAY'
_ENV_TIMEOUT = f'{_ENV_PREFIX}TIMEOUT'
_ENV_CONCURRENCY = f'{_ENV_PREFIX}CONCURRENCY'
_ENV_CONCURRENT_EXECUTOR = f'{_ENV_PREFIX}CONCURRENT_EXECUTOR'
_ENV_REPORT = f'{_ENV_PREFIX}REPORT'
_ENV_PLUGIN = f'{_ENV_PREFIX}PLUGIN'


@command()
@argument('paths', metavar='path', nargs=-1, type=Path(exists=True))
@option(
    'base_url',
    '-u',
    '--base-url',
    help='specify the base URL',
    envvar=_ENV_BASE_URL,
    default='',
)
@option(
    'arguments',
    '-a',
    '--argument',
    help='scenario arguments in format "NAME=VALUE"',
    type=ArgumentType(),
    envvar=_ENV_ARGUMENT,
    multiple=True,
    callback=pairs_callback,
)
@option(
    'level',
    '-l',
    '--level',
    help='show only above or equal to this level',
    type=LevelType(),
    envvar=_ENV_LEVEL,
    default='success',
)
@option(
    'report_dir',
    '-R',
    '--report',
    help='set the report directory',
    type=Path(file_okay=False, writable=True),
    envvar=_ENV_REPORT,
)
@option(
    'retry',
    '-r',
    '--retry',
    help='set the max retry count',
    metavar='num',
    type=IntRange(min=0),
    envvar=_ENV_RETRY,
    default=0,
)
@option(
    'delay',
    '-d',
    '--delay',
    help='set the delay between attempts in seconds',
    metavar='sec',
    type=FloatRange(min=0.0),
    envvar=_ENV_DELAY,
    default=0.1,
)
@option(
    'timeout',
    '-t',
    '--timeout',
    help='set the delay between attempts in seconds',
    metavar='sec',
    type=FloatRange(min=0.0),
    envvar=_ENV_TIMEOUT,
    callback=positive_float_callback,
)
@option(
    'concurrency',
    '-c',
    '--concurrency',
    help='set the concurrency',
    metavar='num',
    type=IntRange(min=1),
    envvar=_ENV_CONCURRENCY,
    default=1,
)
@option(
    'executor_factory',
    '-E',
    '--executor',
    help='set the concurrent executor',
    type=ExecutorFactoryType(),
    envvar=_ENV_CONCURRENT_EXECUTOR,
    default='process',
)
@option(
    'plugins',
    '-p',
    '--plugin',
    help='add a plugin',
    metavar='path',
    type=Path(exists=True),
    multiple=True,
    envvar=_ENV_PLUGIN,
)
@option(
    'verbosity',
    '-v',
    '--verbose',
    help='make logging more verbose',
    count=True,
)
@help_option('-h', '--help')
@version_option(_version)
def main(
    paths: Sequence[str],
    base_url: str,
    arguments: Arguments,
    level: Status,
    report_dir: Optional[str],
    retry: int,
    delay: float,
    timeout: Optional[float],
    concurrency: int,
    executor_factory: Callable[[int], Executor],
    plugins: Iterable[str],
    verbosity: int,
) -> None:
    """Preacher CLI: Web API Verification without Coding"""
    exit_code = app(
        paths=paths,
        base_url=base_url,
        arguments=arguments,
        level=level,
        report_dir=report_dir,
        retry=retry,
        delay=delay,
        timeout=timeout,
        concurrency=concurrency,
        executor_factory=executor_factory,
        plugins=plugins,
        verbosity=verbosity,
    )
    sys.exit(exit_code)



"""An executor."""

from functools import partial
from typing import Optional, Tuple

import requests

from preacher.core.request import Request, Response, Requester, ExecutionReport
from preacher.core.scenario.util.retry import retry_while_false
from preacher.core.value import ValueContext
from preacher.core.verification import ResponseDescription, ResponseVerification

Result = Tuple[ExecutionReport, Optional[Response], Optional[ResponseVerification]]


def predicate(result: Result) -> bool:
    execution, _, verification = result
    if not execution.status.is_succeeded:
        return False
    if verification is None:
        return True
    return verification.status.is_succeeded


class UnitRunner:

    def __init__(self, requester: Requester, retry: int = 0, delay: float = 0.1):
        if retry < 0:
            raise ValueError(f'`retry` must be zero or positive, given {retry}')

        self._requester = requester
        self._retry = retry
        self._delay = delay

    @property
    def base_url(self) -> str:
        return self._requester.base_url

    def run(
        self,
        request: Request,
        requirements: ResponseDescription,
        session: Optional[requests.Session] = None,
    ) -> Result:
        return retry_while_false(
            partial(self._execute, request, requirements, session),
            attempts=self._retry + 1,
            delay=self._delay,
            predicate=predicate,
        )

    def _execute(
        self,
        request: Request,
        requirements: ResponseDescription,
        session: Optional[requests.Session] = None,
    ) -> Result:
        execution, response = self._requester.execute(request, session=session)
        if not response:
            return execution, None, None

        context = ValueContext(origin_datetime=execution.starts)
        verification = requirements.verify(response, context)
        return execution, response, verification



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

import requests

from preacher.core.datetime import now
from preacher.core.extraction import analyze_data_obj
from preacher.core.unit import UnitRunner
from preacher.core.value import ValueContext
from preacher.core.verification import Verification
from .case import Case
from .case_listener import CaseListener
from .case_result import CaseResult


@dataclass(frozen=True)
class CaseContext:
    starts: datetime = field(default_factory=now)
    base_url: str = ''


class CaseRunner:

    def __init__(self, unit_runner: UnitRunner, listener: Optional[CaseListener] = None):
        self._unit_runner = unit_runner
        self._listener = listener or CaseListener()

    @property
    def base_url(self) -> str:
        return self._unit_runner.base_url

    def run(self, case: Case, session: Optional[requests.Session] = None) -> CaseResult:
        if not case.enabled:
            return CaseResult(label=case.label)

        context = CaseContext(base_url=self._unit_runner.base_url)
        context_analyzer = analyze_data_obj(context)
        value_context = ValueContext(origin_datetime=context.starts)
        conditions = Verification.collect(
            condition.verify(context_analyzer, value_context)
            for condition in case.conditions
        )
        if not conditions.status.is_succeeded:
            return CaseResult(case.label, conditions)

        execution, response, verification = self._unit_runner.run(
            request=case.request,
            requirements=case.response,
            session=session,
        )
        self._listener.on_execution(execution, response)

        return CaseResult(case.label, conditions, execution, verification)



"""
YAML handling.
"""

from .error import YamlError
from .loader import load, load_all, load_from_path, load_all_from_path

__all__ = [
    'YamlError',
    'load',
    'load_from_path',
    'load_all',
    'load_all_from_path',
]



from __future__ import annotations

from typing import List, Optional

from preacher.core.request import Response, ExecutionReport
from preacher.core.scenario import ScenarioResult
from preacher.core.scheduling import Listener
from preacher.core.status import Status
from preacher.presentation.html import HtmlReporter


class HtmlReportingListener(Listener):

    def __init__(self, reporter: HtmlReporter):
        self._reporter = reporter
        self._results: List[ScenarioResult] = []

    def on_execution(self, execution: ExecutionReport, response: Optional[Response]) -> None:
        if not response:
            return
        self._reporter.export_response(execution, response)

    def on_scenario(self, result: ScenarioResult) -> None:
        self._results.append(result)

    def on_end(self, status: Status) -> None:
        self._reporter.export_results(self._results)

    @staticmethod
    def from_path(path: str) -> HtmlReportingListener:
        reporter = HtmlReporter(path)
        return HtmlReportingListener(reporter)



from __future__ import annotations

import logging

from preacher.core.scenario import ScenarioResult
from preacher.core.scheduling import Listener
from preacher.core.status import Status
from preacher.presentation.logging import LoggingReporter


class LoggingReportingListener(Listener):

    def __init__(self, reporter: LoggingReporter):
        self._reporter = reporter

    def on_scenario(self, result: ScenarioResult) -> None:
        self._reporter.show_scenario_result(result)

    def on_end(self, status: Status) -> None:
        self._reporter.show_status(status)

    @staticmethod
    def from_logger(logger: logging.Logger) -> LoggingReportingListener:
        reporter = LoggingReporter(logger)
        return LoggingReportingListener(reporter)
